/**********************************************************************
 *
 * mgsearch.h -- 
 * Copyright (C) 1999  The New Zealand Digital Library Project
 *
 * PUT COPYRIGHT NOTICE HERE
 *
 * $Id: mgsearch.h 334 1999-07-01 09:29:21Z rjmcnab $
 *
 *********************************************************************/


#ifndef MGSEARCH_H
#define MGSEARCH_H


#include "text_t.h"
#include "querycache.h"

#define RESULTCACHESIZE 10
#define MAXNUMDOCS      500
#define MAXNUMTERMS     100

class mgsearchclass {
public:
  mgsearchclass ();
  virtual ~mgsearchclass ();

  // the index directory must be set before any searching
  // is done
  void setcollectdir (const text_t &thecollectdir);

  // you only need to use this function before doing any stemming
  // casefolding and stemming will be set if values for them are
  // provided (0 or 1).
  // makeindexcurrent returns true if it was able to load the database
  bool makeindexcurrent (const text_t &index, const text_t &collection,
			 int casefolding = -1, int stemming = -1);

  // stem word uses the values set in the last call to makeindexcurrent
  // to stem the word. It is assumed that word is in unicode
  text_t stemword (const text_t &word);
  text_t stemword (text_t::const_iterator here, text_t::const_iterator end);
  
  // the search results are returned in queryresults
  // search returns 'true' if it was able to do a search
  bool search(const queryparamclass &queryparams, 
	      queryresultsclass &queryresults);
  
  // the document text for 'docnum' is placed in 'output'
  // docTargetDocument returns 'true' if it was able to
  // try to get a document
  // collection is needed to see if an index from the 
  // collection is loaded. If no index has been loaded
  // defaultindex is needed to load one
  bool docTargetDocument(const text_t &defaultindex, 
			 const text_t &collection,
			 int docnum,
			 text_t &output);

  // same as docTargetDocument except the document is returned
  // in its raw form. UDoc remains the property of mg and may
  // be destroyed on the next call to mg.
  bool mgdocument (const text_t &defaultindex, 
		   const text_t &collection,
		   int docnum,
		   char *&UDoc, int &ULen);

protected:
  querycache *cache;
  text_t collectdir; // the collection directory

  void setsearchmode (const queryparamclass &queryparams);
  void submitquery (const queryparamclass &queryparams);
  void getresults (const queryparamclass &queryparams, queryresultsclass &queryresults);

  virtual void extractquoted (text_t &ttquerystring, text_t &ttquotedquery);
  virtual void filterquery (text_t &ttquerystring);
};


#endif
