/**********************************************************************
 *
 * formattools.cpp -- 
 * Copyright (C) 1999  The New Zealand Digital Library Project
 *
 * PUT COPYRIGHT NOTICE HERE
 *
 * $Id: formattools.cpp 347 1999-07-07 05:49:35Z sjboddie $
 *
 *********************************************************************/

/*
   $Log$
   Revision 1.1  1999/07/07 05:49:34  sjboddie
   had another crack at the format string code - created a new formattools
   module. It can now handle {If} and {Or} statements although there's a
   bug preventing nested if's and or's.

 */


#include "formattools.h"


void metadata_t::clear() {
  metaindex = 0;
  parentcommand = pNone;
  parentindex = 0;
  parentoptions.clear();
};

void decision_t::clear() {
  command = dMeta;
  meta.clear();
};


void format_t::clear() {
  command = comText;
  decision.clear();
  text.clear();
  meta.clear();
  nextptr = NULL;
  ifptr = NULL;
  elseptr = NULL;
  orptr = NULL;
};


bool parse_string (const text_t &formatstring, format_t *formatlistptr, 
		   text_tarray &metadata, bool &getParents, 
		   text_tmap &metamap, int &metacount) {

  text_t text;
  text_t::const_iterator here = formatstring.begin();
  text_t::const_iterator end = formatstring.end();

  while (here != end) {

    if (*here == '\\')
      text.push_back (*(++here));

    else if (*here == '{') {
      if (!text.empty()) {
	formatlistptr->command = comText;
	formatlistptr->text = text;
	formatlistptr->nextptr = new format_t();
	formatlistptr = formatlistptr->nextptr;
	
	text.clear();
      }
      if (parse_action (++here, end, formatlistptr, metadata, 
			getParents, metamap, metacount)) {
	formatlistptr->nextptr = new format_t();
	formatlistptr = formatlistptr->nextptr;
	if (here == end) break;
      }
    } else if (*here == '[') {
      if (!text.empty()) {
	formatlistptr->command = comText;
	formatlistptr->text = text;
	formatlistptr->nextptr = new format_t();
	formatlistptr = formatlistptr->nextptr;

	text.clear();
      }
      text_t meta;
      here ++;
      while (*here != ']') {
	if (here == end) return false;
	meta.push_back (*here);
	here ++;
      }
      if (parse_meta (meta, metacount, formatlistptr, metadata, getParents, metamap)) {
	formatlistptr->nextptr = new format_t();
	formatlistptr = formatlistptr->nextptr;
      }

    } else 
      text.push_back (*here);

    here ++;
  }
  if (!text.empty()) {
    formatlistptr->command = comText;
    formatlistptr->text = text;
    formatlistptr->nextptr = new format_t();
    formatlistptr = formatlistptr->nextptr;

  }
  return true;
}

bool parse_action (text_t::const_iterator &here, const text_t::const_iterator &end, 
		   format_t *formatlistptr, text_tarray &metadata, bool &getParents,
		   text_tmap &metamap, int &metacount) {

  text_t::const_iterator it = findchar (here, end, '}');
  if (it == end) return false;

  text_t com = substr (here, it);
  here = findchar (it, end, '{');
  if (here == end) return false;
  else here ++;

  if (com == "If") formatlistptr->command = comIf;
  else if (com == "Or") formatlistptr->command = comOr;
  else return false;

  int curlycount = 0;
  int commacount = 0;
  text_t text;
  while (here != end) {
    if (*here == '{') {curlycount ++; text.push_back(*here);}
    else if (*here == '}' && curlycount > 0) {
      curlycount --; 
      text.push_back(*here);
    }

    else if ((*here == ',' || *here == '}') && curlycount <= 0) {

      if (formatlistptr->command == comOr) {
	// the {Or}{this, or this, or this, or this} statement
	// or'ed statements may be either [metadata] or plain text
	format_t *or_ptr;
	
	// find the next unused orptr
	if (formatlistptr->orptr == NULL) {
	  formatlistptr->orptr = new format_t();
	  or_ptr = formatlistptr->orptr;
	} else {
	  or_ptr = formatlistptr->orptr;
	  while (or_ptr->nextptr != NULL)
	    or_ptr = or_ptr->nextptr;
	  or_ptr->nextptr = new format_t();
	  or_ptr = or_ptr->nextptr;
	}

	text_t::const_iterator beginbracket = text.begin();
	text_t::const_iterator endbracket = (text.end() - 1);
	if ((*beginbracket == '[') && (*endbracket == ']')) {
	  // it's metadata
	  text_t meta = substr (beginbracket+1, endbracket);
	  parse_meta (meta, metacount, or_ptr, metadata, getParents, metamap);

	} else {
	  // assume it's plain text
	  or_ptr->command = comText;
	  or_ptr->text = text;
	}
	text.clear();

      } else {
	// the {If}{decide,do,else} statement
	if (commacount == 0) {
	  // If decision only supports metadata at present
	  
	  // remove the surrounding square brackets
	  text_t::const_iterator beginbracket = text.begin();
	  text_t::const_iterator endbracket = (text.end() - 1);
	  if ((*beginbracket == '[') && (*endbracket == ']')) {
	    text_t meta = substr (beginbracket+1, endbracket);
	    decision_t decision;
	    parse_meta (meta, metacount, formatlistptr->decision, 
			metadata, getParents, metamap);
	    commacount ++;
	    text.clear();
	  } 
	  
	} else if (commacount == 1) {
	  formatlistptr->ifptr = new format_t();
	  parse_string (text, formatlistptr->ifptr, metadata, 
			getParents, metamap, metacount);
	  commacount ++;
	  text.clear();
	  
	} else if (commacount == 2) {
	  formatlistptr->elseptr = new format_t();
	  parse_string (text, formatlistptr->elseptr, metadata, 
			getParents, metamap, metacount);
	  commacount ++;
	  text.clear();
	}
      }
      if (*here == '}') break;
    }

    else 
      text.push_back(*here);
    
    here ++;
  }

  return true;
}

bool parse_formatstring (const text_t &formatstring, format_t *formatlistptr, 
			 text_tarray &metadata, bool &getParents) {

  formatlistptr->clear();
  metadata.erase (metadata.begin(), metadata.end());
  getParents = false;

  text_tmap metamap;
  int metacount = 0;
  return (parse_string (formatstring, formatlistptr, metadata, 
			getParents, metamap, metacount));
}

bool parse_meta (text_t &meta, int &count, decision_t &decision,
		 text_tarray &metadata, bool &getParents, text_tmap &metamap) {

  if (meta.size() > 7 && (substr (meta.begin(), meta.begin()+6) == "parent")) {
    getParents = true;
    metadata_t metaoption;
    get_parent_options (meta, metaoption);
    decision.meta = metaoption;
  }
  text_tmap::const_iterator it;
  if ((it = metamap.find(meta)) != metamap.end()) {
      decision.meta.metaindex = (*it).second.getint();
  } else {
    metamap[meta] = count;
    metadata.push_back (meta);
    decision.meta.metaindex = count;
    count ++;
  }
  return true;
}

bool parse_meta (text_t &meta, int &count, format_t *formatlistptr,
		 text_tarray &metadata, bool &getParents, text_tmap &metamap) {
  
  if (meta == "link")
    formatlistptr->command = comLink;
  else if (meta == "/link")
    formatlistptr->command = comEndLink;

  else {
    formatlistptr->command = comMeta;
    if (meta.size() > 7 && (substr (meta.begin(), meta.begin()+6) == "parent")) {
      getParents = true;
      metadata_t metaoption;
      get_parent_options (meta, metaoption);
      formatlistptr->meta = metaoption;
    }
    text_tmap::const_iterator it;
    if ((it = metamap.find(meta)) != metamap.end()) {
      (formatlistptr->meta).metaindex = (*it).second.getint();
    } else {
      metamap[meta] = count;
      metadata.push_back (meta);
      (formatlistptr->meta).metaindex = count;
      count ++;
    }
  }
  return true;
}

void get_parent_options (text_t &instring, metadata_t &metaoption) {

  metaoption.clear();
  text_t meta, com, op;
  bool inbraces = false;
  bool inquotes = false;
  bool foundcolon = false;
  text_t::const_iterator here = instring.begin()+6;
  text_t::const_iterator end = instring.end();
  while (here != end) {
    if (*here == '(') inbraces = true;
    else if (*here == ')') inbraces = false;
    else if (*here == '\'' && !inquotes) inquotes = true;
    else if (*here == '\'' && inquotes) inquotes = false;
    else if (*here == ':' && !inbraces) foundcolon = true;
    else if (foundcolon) meta.push_back (*here);
    else if (inquotes) op.push_back (*here);
    else com.push_back (*here);
    here ++;
  }
  instring = meta;
  if (com.empty())
    metaoption.parentcommand = pImmediate;
  else if (com == "Top")
    metaoption.parentcommand = pTop;
  else if (is_number(com)) {
    metaoption.parentcommand = pIndex;
    metaoption.metaindex = com.getint();
  } else if (com == "All") {
    metaoption.parentcommand = pAll;
    metaoption.parentoptions = op;
  }
}

text_t get_meta (const ResultDocInfo_t &docinfo, const metadata_t &meta) {
  int metasize = docinfo.metadata.size();
  int mindex = meta.metaindex;
  if (metasize < 1 || metasize <= mindex) return "";
  int valuesize = docinfo.metadata[mindex].values.size();

  switch (meta.parentcommand) {
  case pNone:
    return docinfo.metadata[mindex].values.back();

  case pImmediate:
    if (valuesize > 1)
      return docinfo.metadata[mindex].values[metasize-2];
    break;

  case pTop:
    if (valuesize > 1)
      return docinfo.metadata[mindex].values[0];
    break;

  case pIndex:
    if (valuesize > meta.parentindex)
      return docinfo.metadata[mindex].values[meta.parentindex];
    break;

  case pAll:
    bool first = true;
    text_t tmp;
    text_tarray::const_iterator here = docinfo.metadata[mindex].values.begin();
    text_tarray::const_iterator end = docinfo.metadata[mindex].values.end();
    while (here != end) {
      if (!first) tmp += meta.parentoptions;
      tmp += *here;
      first = false;
      here ++;
    }
    return tmp;
  }
  return "";
}


text_t get_formatted_string (const ResultDocInfo_t &docinfo, format_t *formatlistptr) {

  text_t ft;
  while (formatlistptr != NULL) {
    ft += format_string (docinfo, formatlistptr);
    formatlistptr = formatlistptr->nextptr;
  }
  return ft;
}

text_t format_string (const ResultDocInfo_t &docinfo, format_t *formatlistptr) {

  if (formatlistptr == NULL) return "";

  switch (formatlistptr->command) {
  case comText:
    return formatlistptr->text;
  case comLink:
    return "<a href=\"_httpdocument_&cl=search&d=" + docinfo.OID + "\">";
  case comEndLink:
    return "</a>";
  case comMeta:
    return get_meta (docinfo, formatlistptr->meta);
  case comIf:
    return get_if (docinfo, formatlistptr->decision, formatlistptr->ifptr, formatlistptr->elseptr);
  case comOr:
    return get_or (docinfo, formatlistptr->orptr);
  }
  return "";
}


text_t get_or (const ResultDocInfo_t &docinfo, format_t *orptr) {

  text_t tmp;
  while (orptr != NULL) {

    tmp = format_string (docinfo, orptr);
    if (!tmp.empty()) return tmp;

    orptr = orptr->nextptr;
  }
  return "";
}

text_t get_if (const ResultDocInfo_t &docinfo, const decision_t &decision,
	       format_t *ifptr, format_t *elseptr) {

  // not much of a choice yet ...
  if (decision.command == dMeta) {
    if (get_meta (docinfo, decision.meta) != "") {
      if (ifptr != NULL)
	return get_formatted_string (docinfo, ifptr);
    }
    else {
      if (elseptr != NULL)
	return get_formatted_string (docinfo, elseptr);
    }
  }
  return "";
}

